
AbstractMIDISocket {
	var <parent, <destination;
	
	*new { arg chan, destination ... args;
		^super.new.prInit(chan.asChannelIndex, destination).performList(\init, args)
	}
	
	init { }		// your class should override this
	
	clear { }		// and this

	prInit { arg ch, dest;
		destination = dest;
		parent = MIDIPort.at(ch) ?? { MIDIChannel.new(ch) };
		parent.add(this);
	}

	free {
//"\n\nFreeing MIDI socket".debug;
//this.dump;
//this.dumpBackTrace;
		parent.remove(this);
		this.clear;
	}
	
	// your class must also implement:
	// destination  (returns the thing being played by this socket, or this if appropriate)
	// active  (if destination is this: return true if the responder is still valid)
	// noteOn  (take action)
	// noteOff (take action)
	
}

BasicMIDISocket : AbstractMIDISocket {
	var <>onPlayer, <>offPlayer;
	
	init { arg offfunc;
		onPlayer = destination;  // since destination is this, I simplify the usual syntax:
		destination = this;	  // new(channel, dest, args)  -- so you can say instead:
		offPlayer = offfunc;	  // new(channel, on_func, off_func)
	}

	clear {
		onPlayer = offPlayer = nil;
	}
	
	active { ^onPlayer.notNil }	// since I'm the destination, I must say if I'm active or not
	
	noteOn { arg note, vel; onPlayer.value(note, vel) }
	
	noteOff { arg note, vel; offPlayer.value(note, vel) }
	
}

MIDIRecSocket : AbstractMIDISocket {
	// allows you to record midi note streams into a buffer (client-side)
	// controllers are ignored at present, maybe I'll add that later
	// this object should exist in the midi hierarchy only while it's recording
	// when you create it, it's on
	// when you do .stopRecord, the socket is freed

	var	<active,	// am I recording?
		<buf,
		<unresolvedNotes,	// noteOn places the note here; noteOff resolves it and puts it into buf
		<clock,
		<parse = false,	// by default, output a raw buffer
						// set to true for a hierarchized buffer (w/ grace and chord notes)
						// can be set as an arg to initRecord
		<>startTime,	// time when recording was commenced
		<>lastNoteTime,
		<moreResponders;	// other objects may want to use the timing data gathered here
	
	init {
		clock = destination ? TempoClock.default;
		destination = this;
		active = true;
		moreResponders = [];
		this.initRecord;
	}
	
	clear {
		active = false;
		moreResponders.do({ |resp|
			resp.midiCleanup;
		});
	}
	
	addResponder { |... responders|
		moreResponders = moreResponders.addAll(responders);
	}
	
	removeResponder { |... responders|
		moreResponders.removeAll(responders);
	}
	
	removeAllResponders { moreResponders = []; }
	
	initRecord { arg argParse;
//"MIDIRecSocket-initRecord".postln;
		parse = argParse ? parse;	// change parse flag if the arg is supplied
		buf = MIDIRecBuf("buf" ++ Main.elapsedTime.round);
		unresolvedNotes = Array.new;
	}
	
	parse_ { arg argParse;
		startTime.isNil.if({
			parse = argParse ? parse;
		}, {
			Error("Cannot set parse flag while recording MIDI.").throw;
		});
	}
	
	stopRecord {
		var timeTemp, noteToStop;
		(buf.size > 0).if({
			buf.stopRecTime = (timeTemp = clock.elapsedBeats) - startTime;
				// if stop was received before last noteOff, must fill in note lengths
				// for any notes not yet cut off
			buf.notes.last.dur = timeTemp - lastNoteTime;
				// noteToStop only needs freq to release
			noteToStop = SequenceNote(unresolvedNotes.collect({ |n| n.freq }));
			moreResponders.do({ |resp|
				resp.noteOff(noteToStop);
			});
			unresolvedNotes.do({ |note|
				note.length = timeTemp - note.length;
			});
			startTime = nil;
			parse.if({ buf = buf.parse }, { buf });	// need to do this before freeing
												// in case there's sth in moreResponders
		});
		this.free;
		^buf
	}
	
	noteOn { arg num, vel;
		var timeTemp, new;
//		buf.isNil.if({	// if rec was started properly by button, this will not be true
//			parent.disable(this);
//		}, {
			timeTemp = clock.elapsedBeats;
			startTime.isNil.if({ startTime = timeTemp });  // 1st note should be at 0
				// fill in previous note's delta
			(buf.notes.size > 0).if({
				buf.notes.last.dur = timeTemp - lastNoteTime;
			});
			unresolvedNotes = unresolvedNotes.add(new = SequenceNote(num,
					// cannot assign new lastNoteTime until after previous delta is recorded
				0, lastNoteTime = timeTemp, vel/127));
			buf.add(new);
				// use the SequenceNote object, not the raw MIDI data
			moreResponders.do({ |resp| resp.noteOn(new) });
//		});
	}
	
	noteOff { arg num, vel;
		var note;
			// search for first unresolved note, with this note num
		(unresolvedNotes.size > 0).if({
			note = unresolvedNotes.detect({ arg x; x.freq == num });
		});
		note.notNil.if({
			unresolvedNotes.remove(note);
				// length now holds the note on time, so subtraction gives the length
			note.length = clock.elapsedBeats - note.length;
			moreResponders.do({ |resp| resp.noteOff(note) });
		});
//		parse.if({
//			nil
//		});
	}
	
}

MIDIThruSocket : AbstractMIDISocket {
	init {
		destination = destination.asChannelIndex;
	}
	
	noteOn { |num, vel|
		MIDIIn.doNoteOnAction(destination.port, destination.channel, num, vel)
	}
	
	noteOff { |num, vel|
		MIDIIn.doNoteOffAction(destination.port, destination.channel, num, vel)
	}
	
	active { ^(destination.respondsTo(\port) and: { destination.respondsTo(\channel) }) }
}
